use std::collections::HashMap;

#[derive(Clone, Debug)]
struct Directory {
    size: i64,
    children: Vec<String>,
}

fn get_size(dir: Directory, directories: HashMap<String, Directory>) -> i64 {
    let mut size = dir.size;
    if !dir.children.is_empty() {
        size += dir
            .children
            .iter()
            .map(|child_name| {
                if let Some((_dir_name, child_dir)) = directories.get_key_value(child_name) {
                    return get_size(child_dir.clone(), directories.clone());
                }
                0
            })
            .sum::<i64>();
    }
    size
}

pub fn process_part1(input: &str) -> i64 {
    let mut pwd = String::from("");
    let mut directories: HashMap<String, Directory> = HashMap::new();
    let lines: Vec<Vec<&str>> = input
        .lines()
        .map(|line| line.split_whitespace().collect::<Vec<&str>>())
        .collect();
    for line in lines {
        if line[0] == "$" {
            if line[1] == "cd" {
                let arg = line[2].to_string();
                match arg.as_str() {
                    "/" => pwd = arg,
                    ".." => match pwd.as_str() {
                        "/" => println!("Already at root"),
                        _ => pwd.truncate(pwd.rfind("/").unwrap()),
                    },
                    _ => match pwd.as_str() {
                        "/" => pwd.push_str(&arg),
                        _ => pwd.push_str(&("/".to_owned() + &arg)),
                    },
                }
                if !directories.contains_key(&pwd) {
                    directories.insert(
                        pwd.clone(),
                        Directory {
                            size: 0,
                            children: vec![],
                        },
                    );
                }
            }
        } else if line[0] == "dir" {
            if let Some((dir_name, dir)) = directories.get_key_value(&pwd) {
                let mut new_children = dir.children.clone();
                let mut new_child = "/".to_owned() + line[1];
                if pwd != *"/" {
                    new_child.insert_str(0, &pwd);
                }
                new_children.push(new_child);
                let new_dir = Directory {
                    size: dir.size,
                    children: new_children,
                };
                directories.insert(dir_name.to_string(), new_dir);
            }
        } else if line[0] != "dir" {
            if let Some((dir_name, dir)) = directories.get_key_value(&pwd) {
                let new_dir = Directory {
                    size: dir.size + line[0].parse::<i64>().unwrap(),
                    children: dir.children.clone(),
                };
                directories.insert(dir_name.to_string(), new_dir);
            }
        }
    }
    let sizes = directories
        .values()
        .map(|directory| get_size(directory.clone(), directories.clone()));

    sizes.filter(|size| size <= &100000).sum()
}

pub fn process_part2(input: &str) -> i64 {
    const TOTAL: i64 = 70000000;
    const MIN_UNUSED: i64 = 30000000;
    let mut pwd = String::from("");
    let mut directories: HashMap<String, Directory> = HashMap::new();
    let lines: Vec<Vec<&str>> = input
        .lines()
        .map(|line| line.split_whitespace().collect::<Vec<&str>>())
        .collect();
    for line in lines {
        if line[0] == "$" {
            if line[1] == "cd" {
                let arg = line[2].to_string();
                match arg.as_str() {
                    "/" => pwd = arg,
                    ".." => match pwd.as_str() {
                        "/" => println!("Already at root"),
                        _ => pwd.truncate(pwd.rfind("/").unwrap()),
                    },
                    _ => match pwd.as_str() {
                        "/" => pwd.push_str(&arg),
                        _ => pwd.push_str(&("/".to_owned() + &arg)),
                    },
                }
                if !directories.contains_key(&pwd) {
                    directories.insert(
                        pwd.clone(),
                        Directory {
                            size: 0,
                            children: vec![],
                        },
                    );
                }
            }
        } else if line[0] == "dir" {
            if let Some((dir_name, dir)) = directories.get_key_value(&pwd) {
                let mut new_children = dir.children.clone();
                let mut new_child = "/".to_owned() + line[1];
                if pwd != *"/" {
                    new_child.insert_str(0, &pwd);
                }
                new_children.push(new_child);
                let new_dir = Directory {
                    size: dir.size,
                    children: new_children,
                };
                directories.insert(dir_name.to_string(), new_dir);
            }
        } else if line[0] != "dir" {
            if let Some((dir_name, dir)) = directories.get_key_value(&pwd) {
                let new_dir = Directory {
                    size: dir.size + line[0].parse::<i64>().unwrap(),
                    children: dir.children.clone(),
                };
                directories.insert(dir_name.to_string(), new_dir);
            }
        }
    }
    let sizes = directories
        .values()
        .map(|directory| get_size(directory.clone(), directories.clone()));
    let unused = TOTAL - sizes.clone().max().unwrap();
    let mut sizes = sizes.collect::<Vec<i64>>();
    sizes.sort();
    for size in sizes {
        if size >= (MIN_UNUSED - unused) {
            return size;
        }
    }
    0
}

#[cfg(test)]
mod tests {
    use super::*;

    const INPUT: &str = "$ cd /
$ ls
dir a
14848514 b.txt
8504156 c.dat
dir d
$ cd a
$ ls
dir e
29116 f
2557 g
62596 h.lst
$ cd e
$ ls
584 i
$ cd ..
$ cd ..
$ cd d
$ ls
4060174 j
8033020 d.log
5626152 d.ext
7214296 k";

    #[test]
    fn part1() {
        let result = process_part1(INPUT);
        assert_eq!(result, 95437);
    }

    #[test]
    fn part2() {
        let result = process_part2(INPUT);
        assert_eq!(result, 24933642);
    }
}
