use std::collections::HashSet;

fn is_visible_from_line(tree_line: Vec<u32>, tree_height: u32) -> bool {
    for &tree in tree_line.iter() {
        if tree >= tree_height {
            return false;
        }
    }
    true
}

fn is_tree_visible(tree_grid: Vec<Vec<u32>>, tree_pos: (usize, usize), tree_height: u32) -> bool {
    let horizontal_line = tree_grid[tree_pos.0].clone();
    let vertical_line: Vec<u32> = tree_grid
        .iter()
        .map(|horizontal_line| horizontal_line[tree_pos.1])
        .collect();

    let (left, right) = vertical_line.split_at(tree_pos.0);
    let (top, bottom) = horizontal_line.split_at(tree_pos.1);
    is_visible_from_line(left.to_vec(), tree_height)
        || is_visible_from_line(right[1..].to_vec(), tree_height)
        || is_visible_from_line(top.to_vec(), tree_height)
        || is_visible_from_line(bottom[1..].to_vec(), tree_height)
}

pub fn process_part1(input: &str) -> u32 {
    let tree_grid = input
        .lines()
        .map(|line| {
            line.chars()
                .map(|char| char.to_digit(10).unwrap())
                .collect()
        })
        .collect::<Vec<Vec<u32>>>();
    let dimension = (tree_grid.len(), tree_grid[0].len());
    let outer_trees = (dimension.0 as u32 * 2) + (dimension.1 as u32 * 2) - 4;
    let mut visible_trees: HashSet<(usize, usize)> = HashSet::new();
    for x in 1..dimension.0 - 1 {
        for y in 1..dimension.1 - 1 {
            if is_tree_visible(tree_grid.clone(), (x, y), tree_grid[x][y]) {
                visible_trees.insert((x, y));
            }
        }
    }
    outer_trees + visible_trees.len() as u32
}

fn seeing_distance(tree_line: Vec<u32>, tree_height: u32) -> u32 {
    let mut count = 0;
    for tree in tree_line {
        count += 1;
        if tree >= tree_height {
            break;
        }
    }
    count
}

fn calc_score(tree_grid: Vec<Vec<u32>>, tree_pos: (usize, usize), tree_height: u32) -> u32 {
    let horizontal_line = tree_grid[tree_pos.0].clone();
    let vertical_line: Vec<u32> = tree_grid
        .iter()
        .map(|horizontal_line| horizontal_line[tree_pos.1])
        .collect();

    let (left, right) = vertical_line.split_at(tree_pos.0);
    let mut left = left.to_vec();
    left.reverse();
    let (top, bottom) = horizontal_line.split_at(tree_pos.1);
    let mut top = top.to_vec();
    top.reverse();
    seeing_distance(left.to_vec(), tree_height)
        * seeing_distance(right[1..].to_vec(), tree_height)
        * seeing_distance(top.to_vec(), tree_height)
        * seeing_distance(bottom[1..].to_vec(), tree_height)
}

pub fn process_part2(input: &str) -> u32 {
    let tree_grid = input
        .lines()
        .map(|line| {
            line.chars()
                .map(|char| char.to_digit(10).unwrap())
                .collect()
        })
        .collect::<Vec<Vec<u32>>>();
    let dimension = (tree_grid.len(), tree_grid[0].len());
    let mut highest_score = 0;
    for x in 1..dimension.0 - 1 {
        for y in 1..dimension.1 - 1 {
            if is_tree_visible(tree_grid.clone(), (x, y), tree_grid[x][y]) {
                let score = calc_score(tree_grid.clone(), (x, y), tree_grid[x][y]);
                if score > highest_score {
                    highest_score = score;
                }
            }
        }
    }
    highest_score
}

#[cfg(test)]
mod tests {
    use super::*;

    const INPUT: &str = "30373
25512
65332
33549
35390
";

    #[test]
    fn part1() {
        let result = process_part1(INPUT);
        assert_eq!(result, 21);
    }

    #[test]
    fn part2() {
        let result = process_part2(INPUT);
        assert_eq!(result, 8);
    }
}
