use std::collections::HashSet;

#[derive(Clone, Copy, Debug)]
enum Move {
    Up(i32),
    Down(i32),
    Left(i32),
    Right(i32),
}

#[derive(Clone)]
struct Position {
    head: (i32, i32),
    tail: (i32, i32),
    visited_positions: HashSet<(i32, i32)>,
}

impl Position {
    fn new() -> Position {
        Position {
            head: (0, 0),
            tail: (0, 0),
            visited_positions: HashSet::from([(0, 0)]),
        }
    }

    fn is_touching(&self) -> bool {
        if self.head == self.tail {
            return true;
        }
        if self.head.0.abs_diff(self.tail.0) <= 1 && self.head.1.abs_diff(self.tail.1) <= 1 {
            return true;
        }
        false
    }

    fn update(&mut self, movement: Move) {
        match movement {
            Move::Up(distance) => {
                for _ in 1..=distance {
                    self.step(Move::Up(1));
                }
            }
            Move::Down(distance) => {
                for _ in 1..=distance {
                    self.step(Move::Down(1));
                }
            }
            Move::Left(distance) => {
                for _ in 1..=distance {
                    self.step(Move::Left(1));
                }
            }
            Move::Right(distance) => {
                for _ in 1..=distance {
                    self.step(Move::Right(1));
                }
            }
        }
    }

    fn square(a: i32) -> i32 {
        a * a
    }

    fn get_distance(point_a: (i32, i32), point_b: (i32, i32)) -> f64 {
        (Self::square(point_b.0 - point_a.0) as f64 + Self::square(point_b.1 - point_a.1) as f64)
            .sqrt()
    }

    fn move_tail(&mut self, movement: Move) {
        let big_distance = Self::get_distance(self.head, self.tail) > 2f64.sqrt();
        if !&self.is_touching() {
            match movement {
                Move::Up(_) => {
                    self.tail.0 = self.head.0 - 1;
                    if big_distance {
                        self.tail.1 -= 1;
                    } else {
                        self.tail.1 = self.head.1;
                    }
                }
                Move::Down(_) => {
                    self.tail.0 = self.head.0 + 1;
                    if big_distance {
                        self.tail.1 += 1;
                    } else {
                        self.tail.1 = self.head.1;
                    }
                }
                Move::Left(_) => {
                    self.tail.1 = self.head.1 + 1;
                    if big_distance {
                        self.tail.0 += 1;
                    } else {
                        self.tail.0 = self.head.0;
                    }
                }
                Move::Right(_) => {
                    self.tail.1 = self.head.1 - 1;
                    if big_distance {
                        self.tail.0 -= 1;
                    } else {
                        self.tail.0 = self.head.0;
                    }
                }
            }
        }
        self.visited_positions.insert(self.tail);
    }

    fn step(&mut self, movement: Move) {
        match movement {
            Move::Up(distance) => self.head.0 += distance,
            Move::Down(distance) => self.head.0 -= distance,
            Move::Left(distance) => self.head.1 -= distance,
            Move::Right(distance) => self.head.1 += distance,
        }
        self.move_tail(movement);
    }
}

fn parse_line(line: &str) -> Move {
    let mut line = line.split_whitespace();
    let direction = line.next();
    let distance = line.next().unwrap().parse::<i32>().unwrap();
    match direction {
        Some("U") => Move::Up(distance),
        Some("D") => Move::Down(distance),
        Some("L") => Move::Left(distance),
        Some("R") => Move::Right(distance),
        _ => panic!("Shouldn_t happen"),
    }
}

pub fn process_part1(input: &str) -> usize {
    let mut position = Position {
        head: (0, 0),
        tail: (0, 0),
        visited_positions: HashSet::from([(0, 0)]),
    };
    let moves = input.lines().map(parse_line);
    for movement in moves {
        position.update(movement);
    }
    position.visited_positions.len()
}
pub fn process_part2(input: &str) -> usize {
    let mut positions = vec![Position::new(); 10];
    let moves = input.lines().map(parse_line);
    for movement in moves {
        let mut steps = 0;
        if let Move::Up(distance) = movement {
            steps = distance;
        } else if let Move::Down(distance) = movement {
            steps = distance;
        } else if let Move::Left(distance) = movement {
            steps = distance;
        } else if let Move::Right(distance) = movement {
            steps = distance;
        }
        for _ in 0..steps {
            for (idx, position) in positions.clone().iter().enumerate() {
                let mut position = position.clone();
                if idx == 0 {
                    match movement {
                        Move::Up(_) => {
                            position.step(Move::Up(1));
                        }
                        Move::Down(_) => {
                            position.step(Move::Down(1));
                        }
                        Move::Left(_) => {
                            position.step(Move::Left(1));
                        }
                        Move::Right(_) => {
                            position.step(Move::Right(1));
                        }
                    }
                    println!("head: {:?}", position.head)
                } else {
                    position.head = positions[idx - 1].tail;
                    position.move_tail(movement);
                    println!("knot nr{}: {:?}", idx, position.head);
                }
                if idx == 9 {
                    println!("tail: {:?}", position.tail)
                }
                positions[idx] = position;
            }
        }
    }
    positions[9].visited_positions.len()
}

#[cfg(test)]
mod tests {
    use super::*;

    const INPUT1: &str = "R 4
U 4
L 3
D 1
R 4
D 1
L 5
R 2";

    const INPUT2: &str = "R 5
U 8
L 8
D 3
R 17
D 10
L 25
U 20";

    #[test]
    fn part1() {
        let result = process_part1(INPUT1);
        assert_eq!(result, 13);
    }

    #[test]
    fn part2() {
        let result = process_part2(INPUT2);
        assert_eq!(result, 36);
    }
}
